const MongoClient = require("mongodb").MongoClient;
const ObjectID    = require("mongodb").ObjectID;
const mm          = require('mm')

module.exports = async (dbName, dsn="mongodb://localhost:27017/") => {
    if (!dbName)
        throw new ReferenceError(`db name does not provided`)

    const mongoClient = new MongoClient(dsn, { useNewUrlParser: true });
    const client      = await mongoClient.connect()
    const db          = client.db(dbName)
    const Savable     = mm(db).Savable
    const slice       = mm(db).sliceSavable 

    function getModels(id){
        const SlicedSavable = slice([id, 'user'])

        class User extends SlicedSavable {
            constructor(...params){
                super(...params)
                
                if (this.moneyEvents instanceof Array)
                    return

                if (this.moneyEvents instanceof SlicedSavable)
                    this.moneyEvents = [this.moneyEvents];

                this.moneyEvents = [];
            }

            static get relations(){
                return {
                    moneyEvents: "user"
                }
            }
        }
        SlicedSavable.addClass(User)

        class Event extends SlicedSavable {
            constructor(...params){
                super(...params)

                if (this.moneyEvents instanceof Array)
                    return

                if (this.moneyEvents instanceof SlicedSavable)
                    this.moneyEvents = [this.moneyEvents];

                this.moneyEvents = [];
            }

            get owner(){
                return this.___owner
            }

            get usersSum(){
                return (async  () => {
                    var result = 0;
                    for (let money of this.moneyEvents){
                        result += (await money).amount
                        console.log('MONEY SUM', money.amount, money._id)
                    }
                    return result;
                })()
            }


            get moneyDiff(){
                return (async  () => {
                    return (await this.usersSum) - (this.total || 0);
                })()
            }

            get avg(){
                return this.total/((this.moneyEvents && this.moneyEvents.length) || 1)
            }


            static get relations(){
                return {
                    moneyEvents: "event"
                }
            }
        }
        SlicedSavable.addClass(Event)

        class EventMoney extends SlicedSavable {
            get owner(){
                return this.___owner
            }

            get avgDiff(){
                return (async () =>  this.amount - (await this.event).avg)()
            }

            static get relations(){
                return {
                    user: "moneyEvents",
                    event: "moneyEvents"
                }
            }
        }
        SlicedSavable.addClass(EventMoney)

        return {
            SlicedSavable, User, Event, EventMoney
        }
    }

    return {
        Savable, 
        slice,
        getModels
    }
}
